﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using AForge.Video;
using AForge.Video.DirectShow;
using AForge.Vision.Motion;
using AForge.Imaging;
using AForge.Imaging.Filters;

namespace Background_Filter
{
    public abstract class BackgroundFilterBase
    {
        // frame to hightlight motion
        protected UnmanagedImage motionFrame, currentFrame;

        // temporary buffer used for suppressing noise
        protected UnmanagedImage tempFrame;

        protected Boolean suppressNoise;

        protected int differenceThreshold = 15;

        protected int differenceThresholdNeg = -15;

        protected int frameCount = 0;

        // binary erosion filter for noise surpression
        protected BinaryErosion3x3 erosionFilter = new BinaryErosion3x3();

        protected BinaryDilatation3x3 dilationFilter = new BinaryDilatation3x3();

        public BackgroundFilterBase(Bitmap videoFrame, Boolean sn)
        {
            suppressNoise = sn;
        }
        

        public unsafe abstract void ProcessFrame(Bitmap videoFrame);

        public Bitmap MotionFrame
        {
            get
            {
                lock (this)
                {
                    return motionFrame.ToManagedImage();
                }
            }
        }

        public bool SuppressNoise
        {
            set
            {
                lock (this)
                {
                    suppressNoise = value;

                    // check if temporary frame is not required
                    if ((!suppressNoise) && (tempFrame != null))
                    {
                        tempFrame.Dispose();
                        tempFrame = null;
                    }
                }
            }
        }
    }

    public class TwoFrameDifferencingFilter : BackgroundFilterBase
    {
        private UnmanagedImage previousFrame;

        public TwoFrameDifferencingFilter(Bitmap videoFrame, Boolean sn)
            : base(videoFrame, sn)
        { }

        public unsafe override void ProcessFrame(Bitmap videoFrame)
        {
            BitmapData videoData = videoFrame.LockBits(
                new Rectangle(0, 0, videoFrame.Width, videoFrame.Height),
                ImageLockMode.ReadWrite, videoFrame.PixelFormat);

            UnmanagedImage frame;

            frame = new UnmanagedImage(videoData);
            int width = frame.Width;
            int height = frame.Height;
            frameCount++;
            lock (this)
            {
                if ((suppressNoise) && (tempFrame==null))
                    tempFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);

                if (previousFrame == null)
                {

                    // allocate memory for previous and current frames
                    previousFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);
                    currentFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);

                    // convert source frame to grayscale
                    Grayscale.CommonAlgorithms.BT709.Apply(frame, previousFrame);
               
                    return;
                }


                // convert current image to grayscale
                Grayscale.CommonAlgorithms.BT709.Apply(frame, currentFrame);
           

                // pointers to previous and current frames
                byte* prevFrame = (byte*)previousFrame.ImageData.ToPointer();
                byte* currFrame = (byte*)currentFrame.ImageData.ToPointer();

                // difference value
                int diff;
                int frameSize = currentFrame.Stride * height;
                for (int y = 0; y < previousFrame.Height; y++)
                {
                    for (int x = 0; x < previousFrame.Width; x++)
                    {
                        diff = (int)*currFrame - (int)*prevFrame;

                        // copy current frame to previous
                        *prevFrame++ = *currFrame;

                        // threshold
                        *currFrame = ((diff >= differenceThreshold) || (diff <= differenceThresholdNeg)) ? (byte)255 : (byte)0;

                        currFrame++;
                       
                    }

                    prevFrame += previousFrame.Stride - previousFrame.Width;
                    currFrame += currentFrame.Stride - currentFrame.Width;
                }            

                if (suppressNoise)
                {
                    // suppress noise
                    AForge.SystemTools.CopyUnmanagedMemory(tempFrame.ImageData, currentFrame.ImageData, frameSize);
                    erosionFilter.Apply(tempFrame, currentFrame);
                    AForge.SystemTools.CopyUnmanagedMemory(tempFrame.ImageData, currentFrame.ImageData, frameSize);
                    dilationFilter.Apply(tempFrame, currentFrame);
                }

                motionFrame = currentFrame;
                
           } 

        }
    }

    public class BackgroundSubtractionFilter : BackgroundFilterBase
    {
        private UnmanagedImage backgroundFrame;

        private double backgroundWeight;    // For weighting the current and background frames

        public BackgroundSubtractionFilter(Bitmap videoFrame, Boolean sn, double bw)
            : base(videoFrame, sn)
        {
            backgroundWeight = bw;
        }

        public unsafe override void ProcessFrame(Bitmap videoFrame)
        {
            BitmapData videoData = videoFrame.LockBits(
                new Rectangle(0, 0, videoFrame.Width, videoFrame.Height),
                ImageLockMode.ReadWrite, videoFrame.PixelFormat);

            UnmanagedImage frame;

            frame = new UnmanagedImage(videoData);
            int width = frame.Width;
            int height = frame.Height;
            frameCount++;
            lock (this)
            {
                if ((suppressNoise) && (tempFrame == null))
                    tempFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);

                if (backgroundFrame == null)
                {

                    // allocate memory for background and current frames
                    backgroundFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);
                    currentFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);

                    // initialize the background frame
                    Grayscale.CommonAlgorithms.BT709.Apply(frame, backgroundFrame);

                    return;
                }

                 // convert current image to grayscale
                Grayscale.CommonAlgorithms.BT709.Apply(frame, currentFrame);

                // pointers to background and current frames
                byte* currFrame = (byte*)currentFrame.ImageData.ToPointer();
                byte* backFrame = (byte*)backgroundFrame.ImageData.ToPointer();

                // difference value
                int diff;
                int frameSize = currentFrame.Stride * height;
                for (int y = 0; y < currentFrame.Height; y++)
                {
                    for (int x = 0; x < currentFrame.Width; x++)
                    {
                        diff = (int)*currFrame - (int)*backFrame;        

                        // update background frame
                        *backFrame++ = (byte)(*currFrame * backgroundWeight + (1 - backgroundWeight) * (*backFrame));

                        *currFrame++ = ((diff >= differenceThreshold) || (diff <= differenceThresholdNeg)) ? (byte)255 : (byte)0;

                    }

                    backFrame += backgroundFrame.Stride - backgroundFrame.Width;
                    currFrame += currentFrame.Stride - currentFrame.Width;
                }


                if (suppressNoise)
                {
                    // suppress noise
                    AForge.SystemTools.CopyUnmanagedMemory(tempFrame.ImageData, currentFrame.ImageData, frameSize);
                    erosionFilter.Apply(tempFrame, currentFrame);
                    AForge.SystemTools.CopyUnmanagedMemory(tempFrame.ImageData, currentFrame.ImageData, frameSize);
                    dilationFilter.Apply(tempFrame, currentFrame);
                }

                motionFrame = currentFrame;

            }

        }
    }

    public class BackgroundSubtractionMedianFilter : BackgroundFilterBase
    {
        private UnmanagedImage backgroundFrame;

        private int medianFilterLength;    // The number of previous frames in the median filter

        private int currentFilterLength = 0;

        private byte[,][] dataBuffer;      // For storing previous frames for the median computation

        public BackgroundSubtractionMedianFilter(Bitmap videoFrame, Boolean sn, int fl)
            : base(videoFrame, sn)
        {
            medianFilterLength = fl;

            dataBuffer=new byte[videoFrame.Width,videoFrame.Height][];

            for (int i = 0; i < videoFrame.Width; i++)
                for (int j = 0; j < videoFrame.Height; j++)
                    dataBuffer[i, j] = new byte[medianFilterLength];
        }

        public unsafe override void ProcessFrame(Bitmap videoFrame)
        {
            BitmapData videoData = videoFrame.LockBits(
                new Rectangle(0, 0, videoFrame.Width, videoFrame.Height),
                ImageLockMode.ReadWrite, videoFrame.PixelFormat);

            UnmanagedImage frame;

            frame = new UnmanagedImage(videoData);
            int width = frame.Width;
            int height = frame.Height;
            frameCount++;
            lock (this)
            {
                if ((suppressNoise) && (tempFrame == null))
                    tempFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);

                if (backgroundFrame == null)
                {

                    // allocate memory for background and current frames and the current frame
                    backgroundFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);
                    currentFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);

                    // initialize the background frame
                    Grayscale.CommonAlgorithms.BT709.Apply(frame, backgroundFrame);

                    return;
                }

                // convert current image to grayscale
                Grayscale.CommonAlgorithms.BT709.Apply(frame, currentFrame);

                // pointers to background and current frames
                byte* currFrame = (byte*)currentFrame.ImageData.ToPointer();
                byte* backFrame = (byte*)backgroundFrame.ImageData.ToPointer();

                // difference value
                int diff;
                int frameSize = currentFrame.Stride * height;
                byte[] tempBuffer = new byte[medianFilterLength];
                for (int y = 0; y < currentFrame.Height; y++)
                {
                    for (int x = 0; x < currentFrame.Width; x++)
                    {
                        if (currentFilterLength < (medianFilterLength - 1))
                        {
                            dataBuffer[x, y][currentFilterLength] = (byte)*currFrame;

                            diff = (int)*currFrame - (int)*backFrame;

                            // update background frame
                            *backFrame++ = (byte)(*currFrame);
                        }
                        else
                        {
                            diff = (int)*currFrame - (int)*backFrame;

                            Array.Copy(dataBuffer[x, y], tempBuffer, medianFilterLength);

                            Array.Sort(tempBuffer);

                            *backFrame++ = tempBuffer[medianFilterLength / 2];

                            // Update databuffer

                            for (int j = 1; j < medianFilterLength; j++)
                                dataBuffer[x, y][j - 1] = dataBuffer[x, y][j];

                            dataBuffer[x, y][medianFilterLength - 1] = *currFrame;

                        }

                        *currFrame++ = ((diff >= differenceThreshold) || (diff <= differenceThresholdNeg)) ? (byte)255 : (byte)0;

                    }

                    backFrame += backgroundFrame.Stride - backgroundFrame.Width;
                    currFrame += currentFrame.Stride - currentFrame.Width;
                }

                currentFilterLength++;


                if (suppressNoise)
                {
                    // suppress noise
                    AForge.SystemTools.CopyUnmanagedMemory(tempFrame.ImageData, currentFrame.ImageData, frameSize);
                    erosionFilter.Apply(tempFrame, currentFrame);
                    AForge.SystemTools.CopyUnmanagedMemory(tempFrame.ImageData, currentFrame.ImageData, frameSize);
                    dilationFilter.Apply(tempFrame, currentFrame);
                }

                motionFrame = currentFrame;

            }

        }
    }

    public class GaussianModelFilter : BackgroundFilterBase
    {
        private UnmanagedImage meanFrame, sdFrame;

        private double probabilityThreshold;    // probability decision threshold

        private double logLikelihoodThreshold;    // log of probability decision threshold

        private double logLikelihood(int grey, double mean, double sd)
        {
            return -Math.Log(sd)-0.5*Math.Log(2.0*Math.PI)-((grey-mean)*(grey-mean)/(2*sd*sd));
        }

        public GaussianModelFilter(Bitmap videoFrame, Boolean sn, double pt)
            : base(videoFrame, sn)
        {
            probabilityThreshold = pt;

            logLikelihoodThreshold = Math.Log(probabilityThreshold);
        }

        public unsafe override void ProcessFrame(Bitmap videoFrame)
        {
            BitmapData videoData = videoFrame.LockBits(
                new Rectangle(0, 0, videoFrame.Width, videoFrame.Height),
                ImageLockMode.ReadWrite, videoFrame.PixelFormat);

            UnmanagedImage frame;

            frame = new UnmanagedImage(videoData);
            int width = frame.Width;
            int height = frame.Height;
            frameCount++;
            lock (this)
            {
                if ((suppressNoise) && (tempFrame == null))
                    tempFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);

                if (meanFrame == null)
                {

                    // allocate memory for current frames and the mean and sd frames
                    meanFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);
                    sdFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);
                    currentFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);

                    // initialize the previous and mean frame
                    Grayscale.CommonAlgorithms.BT709.Apply(frame, meanFrame);

                    return;
                }

                // convert current image to grayscale
                Grayscale.CommonAlgorithms.BT709.Apply(frame, currentFrame);


                // pointers to previous, current, mean and sd frames
                byte* currFrame = (byte*)currentFrame.ImageData.ToPointer();
                byte* mFrame = (byte*)meanFrame.ImageData.ToPointer();
                byte* sFrame = (byte*)sdFrame.ImageData.ToPointer();

                double n = Math.Min((double)(frameCount),30.0);

                double alpha = (n - 1) / n;
                alpha = 0.1;
                int frameSize = currentFrame.Stride * height;
                for (int y = 0; y < currentFrame.Height; y++)
                {
                    for (int x = 0; x < currentFrame.Width; x++)
                    {

                        // update mean and sd frames
                      //  *mFrame = (byte)((*mFrame)*alpha+(*currFrame)/n);
                        *mFrame = (byte)((*mFrame) * alpha + (1-alpha)*(*currFrame));

                        double sd = (*sFrame);

                        sd *= sd;
                      //  sd = alpha * sd + (*currFrame - *mFrame) * (*currFrame - *mFrame)/(n-1);
                        sd = alpha * sd + (1-alpha)*(*currFrame - *mFrame) * (*currFrame - *mFrame);

                        *sFrame = (byte)Math.Sqrt(sd);

                        // compute log likelihood
                        double ll = logLikelihood(*currFrame, *mFrame, *sFrame);
                       
                        *currFrame++ = (ll<logLikelihoodThreshold ) ? (byte)255 : (byte)0;

                        mFrame++;
                        sFrame++;

                    }

                    currFrame += currentFrame.Stride - currentFrame.Width;
                    mFrame += meanFrame.Stride - meanFrame.Width;
                    sFrame += sdFrame.Stride - sdFrame.Width;
                }            


                if (suppressNoise)
                {
                    // suppress noise
                    AForge.SystemTools.CopyUnmanagedMemory(tempFrame.ImageData, currentFrame.ImageData, frameSize);
                    erosionFilter.Apply(tempFrame, currentFrame);
                    AForge.SystemTools.CopyUnmanagedMemory(tempFrame.ImageData, currentFrame.ImageData, frameSize);
                    dilationFilter.Apply(tempFrame, currentFrame);
                }

                motionFrame = currentFrame;

            }

        }
    }
    
    /*
    public class RelaxationLabellingFilter : BackgroundFilterBase
    {
        private UnmanagedImage previousFrame;

        private double[,] probabilityFrame;
        
        private double[,] previousProbabilityFrame;

        private double[,] tempProbabilityFrame;

        private double probabilityThreshold;

        private int maxIterationCount;

        public RelaxationLabellingFilter(Bitmap videoFrame, Boolean sn, double pt)
            : base(videoFrame, sn)
        {
            maxIterationCount = 7;

            probabilityThreshold = pt;
        }

        public unsafe override void ProcessFrame(Bitmap videoFrame)
        {
            BitmapData videoData = videoFrame.LockBits(
                new Rectangle(0, 0, videoFrame.Width, videoFrame.Height),
                ImageLockMode.ReadWrite, videoFrame.PixelFormat);

            UnmanagedImage frame;

            frame = new UnmanagedImage(videoData);
            int width = frame.Width;
            int height = frame.Height;
            int diff, frameSize;
            frameCount++;
            byte* prevFrame;
            byte* currFrame;

           
            lock (this)
            {
                if ((suppressNoise) && (tempFrame == null))
                    tempFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);

                if (previousFrame == null)
                {

                    // allocate memory for previous and current frames and the current frame
                    previousFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);
                    currentFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);

                    // convert source frame to grayscale
                    Grayscale.CommonAlgorithms.BT709.Apply(frame, previousFrame);
                }

                if (previousProbabilityFrame == null)
                {
                    // allocate memory for previous and current probaility frames
                    probabilityFrame = new double[videoFrame.Width, videoFrame.Height];

                    previousProbabilityFrame = new double[videoFrame.Width, videoFrame.Height];

                    tempProbabilityFrame = new double[videoFrame.Width, videoFrame.Height];

                    // convert source frame to grayscale
                    Grayscale.CommonAlgorithms.BT709.Apply(frame, currentFrame);

                     // pointers frames
                    prevFrame = (byte*)previousFrame.ImageData.ToPointer();
                    currFrame = (byte*)currentFrame.ImageData.ToPointer();

                    frameSize = currentFrame.Stride * height;
                    for (int y = 0; y < previousFrame.Height; y++)
                    {
                        for (int x = 0; x < previousFrame.Width; x++)
                        {
                            diff = (int)*currFrame - (int)*prevFrame;

                            // copy current frame to previous
                            *prevFrame++ = *currFrame++;

                            // initialize current probability frame
                            previousProbabilityFrame[x, y] = Math.Abs(diff)/255.0;

                        }

                        prevFrame += previousFrame.Stride - previousFrame.Width;
                        currFrame += currentFrame.Stride - currentFrame.Width;
                    }        

                    return;
                }  

                // convert current image to grayscale
                Grayscale.CommonAlgorithms.BT709.Apply(frame, currentFrame);

                prevFrame = (byte*)previousFrame.ImageData.ToPointer();
                currFrame = (byte*)currentFrame.ImageData.ToPointer();
                frameSize = currentFrame.Stride * height;
                for (int y = 0; y < currentFrame.Height; y++)
                {
                    for (int x = 0; x < currentFrame.Width; x++)
                    {
                        diff = (int)*currFrame - (int)*prevFrame;

                        *prevFrame++ = *currFrame++;

                        // initialize current probability frame
                        probabilityFrame[x, y] = Math.Abs(diff) / 255.0;

                    }
                    prevFrame += previousFrame.Stride - previousFrame.Width;
                    currFrame += currentFrame.Stride - currentFrame.Width;
                } 
       
                // perform relaxation iteration
                double dp;
                for (int k = 0; k < maxIterationCount; k++)
                {
                    for (int y = 0; y < currentFrame.Height; y++)
                    {
                        for (int x = 0; x < currentFrame.Width; x++)
                        {

                            if ((y > 0) && (x > 0) && (y < (currentFrame.Height - 1)) && (x < (currentFrame.Width - 1)))
                            {

                                dp = 0;
                                for (int x2 = x - 1; x2 <= x + 1; x2++)
                                    for (int y2 = y - 1; y2 <= y + 1; y2++)
                                    {
                                        if ((x2 != x) || (y2 != y))
                                            dp += ((1.0 - probabilityThreshold) * probabilityFrame[x2, y2] - probabilityThreshold * (1 - probabilityFrame[x2, y2]));
                                    }
                                dp += previousProbabilityFrame[x, y];
                                dp /= 9.0;

                                double d1 = probabilityFrame[x, y] * (1 + dp);
                                double d2 = (1 - probabilityFrame[x, y]) * (1 - dp);

                                tempProbabilityFrame[x, y] = d1 / (d1 + d2);
                            }
                        }
                    }

                    for (int y = 0; y < currentFrame.Height; y++)
                    {
                        for (int x = 0; x < currentFrame.Width; x++)
                        {
                            probabilityFrame[x,y] = tempProbabilityFrame[x, y];
                        }
                    }
                }  

                // copy to current frame
                currFrame = (byte*)currentFrame.ImageData.ToPointer();
                for (int y = 0; y < currentFrame.Height; y++)
                {
                    for (int x = 0; x < currentFrame.Width; x++)
                    {
                        if (probabilityFrame[x, y] > 0.5)
                            *currFrame = 255;
                        else
                            *currFrame = 0;

                        currFrame++;

                        previousProbabilityFrame[x, y] = probabilityFrame[x, y];

                    }

                    currFrame += currentFrame.Stride - currentFrame.Width;
                } 

                if (suppressNoise)
                {
                    // suppress noise
                    AForge.SystemTools.CopyUnmanagedMemory(tempFrame.ImageData, currentFrame.ImageData, frameSize);
                    erosionFilter.Apply(tempFrame, currentFrame);
                    AForge.SystemTools.CopyUnmanagedMemory(tempFrame.ImageData, currentFrame.ImageData, frameSize);
                    dilationFilter.Apply(tempFrame, currentFrame);
                }

                motionFrame = currentFrame;

            }

        }
    } */
     
    

    public class RelaxationLabellingFilter : BackgroundFilterBase
    {
        private UnmanagedImage previousFrame;

        private UnmanagedImage probabilityFrame;

        private UnmanagedImage previousProbabilityFrame;

        private UnmanagedImage tempProbabilityFrame;

        private double probabilityThreshold;

        private int maxIterationCount;

        public RelaxationLabellingFilter(Bitmap videoFrame, Boolean sn, double pt)
            : base(videoFrame, sn)
        {
            maxIterationCount =5;

            probabilityThreshold = pt;
        }

        public unsafe override void ProcessFrame(Bitmap videoFrame)
        {
            BitmapData videoData = videoFrame.LockBits(
                new Rectangle(0, 0, videoFrame.Width, videoFrame.Height),
                ImageLockMode.ReadWrite, videoFrame.PixelFormat);

            UnmanagedImage frame;

            frame = new UnmanagedImage(videoData);
            int width = frame.Width;
            int height = frame.Height;
            int diff, frameSize;
            frameCount++;
            byte* prevFrame;
            byte* currFrame;
            byte* probFrame;
            byte* prevProbFrame;
            byte* tempProbFrame;


            lock (this)
            {
                if ((suppressNoise) && (tempFrame == null))
                    tempFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);

                if (previousFrame == null)
                {

                    // allocate memory for previous and current frames and the current frame
                    previousFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);
                    currentFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);

                    // convert source frame to grayscale
                    Grayscale.CommonAlgorithms.BT709.Apply(frame, previousFrame);
                }

                if (previousProbabilityFrame == null)
                {
                    // allocate memory for previous and current probaility frames
                    probabilityFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);

                    previousProbabilityFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);

                    tempProbabilityFrame = UnmanagedImage.Create(width, height, PixelFormat.Format8bppIndexed);

                    // convert source frame to grayscale
                    Grayscale.CommonAlgorithms.BT709.Apply(frame, currentFrame);

                    // pointers frames
                    prevFrame = (byte*)previousFrame.ImageData.ToPointer();
                    currFrame = (byte*)currentFrame.ImageData.ToPointer();
                    prevProbFrame = (byte*)previousProbabilityFrame.ImageData.ToPointer();

                    frameSize = currentFrame.Stride * height;
                    for (int y = 0; y < previousFrame.Height; y++)
                    {
                        for (int x = 0; x < previousFrame.Width; x++)
                        {
                            diff = (int)*currFrame - (int)*prevFrame;

                            // copy current frame to previous
                            *prevFrame++ = *currFrame++;

                            // initialize current probability frame
                            *prevProbFrame++ = (byte) Math.Abs(diff);

                        }

                        prevFrame += previousFrame.Stride - previousFrame.Width;
                        currFrame += currentFrame.Stride - currentFrame.Width;
                        prevProbFrame += previousProbabilityFrame.Stride - previousProbabilityFrame.Width;
                    }

                    return;
                }

                // convert current image to grayscale
                Grayscale.CommonAlgorithms.BT709.Apply(frame, currentFrame);

                prevFrame = (byte*)previousFrame.ImageData.ToPointer();
                currFrame = (byte*)currentFrame.ImageData.ToPointer();
                probFrame = (byte*)probabilityFrame.ImageData.ToPointer();
                frameSize = currentFrame.Stride * height;
                for (int y = 0; y < currentFrame.Height; y++)
                {
                    for (int x = 0; x < currentFrame.Width; x++)
                    {
                        diff = (int)*currFrame - (int)*prevFrame;

                        *prevFrame++ = *currFrame++;

                        // initialize current probability frame
                        *probFrame++ = (byte)Math.Abs(diff);


                    }
                    prevFrame += previousFrame.Stride - previousFrame.Width;
                    currFrame += currentFrame.Stride - currentFrame.Width;
                    probFrame += probabilityFrame.Stride - probabilityFrame.Width;
                }

                // perform relaxation iteration
                
                double dp;
                for (int k = 0; k < maxIterationCount; k++)
                {
                    probFrame = (byte*)probabilityFrame.ImageData.ToPointer();
                    prevProbFrame = (byte*)previousProbabilityFrame.ImageData.ToPointer();
                    tempProbFrame = (byte*)tempProbabilityFrame.ImageData.ToPointer();
                    for (int y = 0; y < currentFrame.Height; y++)
                    {
                        for (int x = 0; x < currentFrame.Width; x++)
                        {

                            if ((y > 0) && (x > 0) && (y < (currentFrame.Height - 1)) && (x < (currentFrame.Width - 1)))
                            {

                                dp = 0;
                                
                                double pf=(*(probFrame-1))/255.0;
                                dp += ((1.0 - probabilityThreshold) * pf - probabilityThreshold * (1 - pf));
                                pf = (*(probFrame + 1)) / 255.0;
                                dp += ((1.0 - probabilityThreshold) * pf - probabilityThreshold * (1 - pf));
                                pf = (*(probFrame + probabilityFrame.Stride)) / 255.0;
                                dp += ((1.0 - probabilityThreshold) * pf - probabilityThreshold * (1 - pf));
                                pf = (*(probFrame + probabilityFrame.Stride - 1)) / 255.0;
                                dp += ((1.0 - probabilityThreshold) * pf - probabilityThreshold * (1 - pf));
                                pf = (*(probFrame + probabilityFrame.Stride + 1)) / 255.0;
                                dp += ((1.0 - probabilityThreshold) * pf - probabilityThreshold * (1 - pf));
                                pf = (*(probFrame - probabilityFrame.Stride)) / 255.0;
                                dp += ((1.0 - probabilityThreshold) * pf - probabilityThreshold * (1 - pf));
                                pf = (*(probFrame - probabilityFrame.Stride - 1)) / 255.0;
                                dp += ((1.0 - probabilityThreshold) * pf - probabilityThreshold * (1 - pf));
                                pf = (*(probFrame - probabilityFrame.Stride + 1)) / 255.0;
                                dp += ((1.0 - probabilityThreshold) * pf - probabilityThreshold * (1 - pf));

                                
                               // dp += (*prevProbFrame)/255.0;
                                dp /= 8;

                                double d1 = (*probFrame/255.0) * (1 + dp);
                                double d2 = (1 - (*probFrame/255.0)) * (1 - dp);

                                *tempProbFrame = (byte)(255.0 * d1 / (d1 + d2));
                            }

                            probFrame++;
                            tempProbFrame++;
                            prevProbFrame++;
                        }

                        probFrame += probabilityFrame.Stride - probabilityFrame.Width;
                        tempProbFrame += tempProbabilityFrame.Stride - tempProbabilityFrame.Width;
                        prevProbFrame += previousProbabilityFrame.Stride - previousProbabilityFrame.Width;
                    }

                    AForge.SystemTools.CopyUnmanagedMemory(probabilityFrame.ImageData, tempProbabilityFrame.ImageData, frameSize);
                }

                // copy to current frame
                currFrame = (byte*)currentFrame.ImageData.ToPointer();
                probFrame = (byte*)probabilityFrame.ImageData.ToPointer();
                prevProbFrame = (byte*)previousProbabilityFrame.ImageData.ToPointer();
                for (int y = 0; y < currentFrame.Height; y++)
                {
                    for (int x = 0; x < currentFrame.Width; x++)
                    {
                        if (*probFrame > 128)
                            *currFrame = 255;
                        else
                            *currFrame = 0;

                        *prevProbFrame = *probFrame;

                        currFrame++;
                        probFrame++;
                        prevProbFrame++;

                    }

                    currFrame += currentFrame.Stride - currentFrame.Width;
                    probFrame += probabilityFrame.Stride - probabilityFrame.Width;
                    prevProbFrame += previousProbabilityFrame.Stride - previousProbabilityFrame.Width;
                }

                if (suppressNoise)
                {
                    // suppress noise
                    AForge.SystemTools.CopyUnmanagedMemory(tempFrame.ImageData, currentFrame.ImageData, frameSize);
                    erosionFilter.Apply(tempFrame, currentFrame);
                    AForge.SystemTools.CopyUnmanagedMemory(tempFrame.ImageData, currentFrame.ImageData, frameSize);
                    dilationFilter.Apply(tempFrame, currentFrame);
                }

                motionFrame = currentFrame;

            }

        }
    } 
}
